from ..error import IdentifierError
from ..objects.chargeback import Chargeback
from .base import Base


class Chargebacks(Base):
    RESOURCE_ID_PREFIX = "chb_"

    def get_resource_object(self, result):
        return Chargeback(result)

    def get(self, chargeback_id, **params):
        """Verify the chargeback ID and retrieve the chargeback from the API."""
        if not chargeback_id or not chargeback_id.startswith(self.RESOURCE_ID_PREFIX):
            raise IdentifierError(
                f"Invalid chargeback ID: '{chargeback_id}'. "
                f"A chargeback ID should start with '{self.RESOURCE_ID_PREFIX}'."
            )
        return super().get(chargeback_id, **params)



from .base import Base
from .issuer import Issuer
from .list import List


class Method(Base):
    @classmethod
    def get_resource_class(cls, client):
        from ..resources.methods import Methods

        return Methods(client)

    BANCONTACT = "bancontact"
    BANKTRANSFER = "banktransfer"
    BELFIUS = "belfius"
    CREDITCARD = "creditcard"
    DIRECTDEBIT = "directdebit"
    EPS = "eps"
    GIFTCARD = "giftcard"
    GIROPAY = "giropay"
    IDEAL = "ideal"
    INGHOMEPAY = "inghomepay"
    KBC = "kbc"
    KLARNAPAYLATER = "klarnapaylater"
    KLARNASLICEIT = "klarnasliceit"
    MEALVOUCHER = "mealvoucher"
    MISTERCASH = "mistercash"
    MYBANK = "mybank"
    PAYPAL = "paypal"
    PAYSAFECARD = "paysafecard"
    PODIUMCADEAUKAART = "podiumcadeaukaart"
    PRZELEWY24 = "przelewy24"
    SOFORT = "sofort"

    @property
    def description(self):
        return self._get_property("description")

    @property
    def id(self):
        return self._get_property("id")

    @property
    def minimum_amount(self):
        return self._get_property("minimumAmount")

    @property
    def maximum_amount(self):
        return self._get_property("maximumAmount")

    @property
    def pricing(self):
        return self._get_property("pricing")

    @property
    def image_svg(self):
        try:
            images = self._get_property("image")
            return images["svg"]
        except KeyError:
            return None

    @property
    def image_size1x(self):
        try:
            images = self._get_property("image")
            return images["size1x"]
        except KeyError:
            return None

    @property
    def image_size2x(self):
        try:
            images = self._get_property("image")
            return images["size2x"]
        except KeyError:
            return None

    @property
    def issuers(self):
        """Return the list of available issuers for this payment method."""
        issuers = self._get_property("issuers") or []
        result = {
            "_embedded": {
                "issuers": issuers,
            },
            "count": len(issuers),
        }
        return List(result, Issuer)



import warnings

from ..error import APIDeprecationWarning
from .base import Base


class Settlement(Base):
    STATUS_OPEN = "open"
    STATUS_PENDING = "pending"
    STATUS_PAIDOUT = "paidout"
    STATUS_FAILED = "failed"

    @classmethod
    def get_resource_class(cls, client):
        from ..resources.settlements import Settlements

        return Settlements(client)

    @property
    def id(self):
        return self._get_property("id")

    @property
    def reference(self):
        return self._get_property("reference")

    @property
    def created_at(self):
        return self._get_property("createdAt")

    @property
    def settled_at(self):
        return self._get_property("settledAt")

    @property
    def status(self):
        return self._get_property("status")

    @property
    def amount(self):
        return self._get_property("amount")

    @property
    def periods(self):
        return self._get_property("periods")

    @property
    def invoice_id(self):
        warnings.warn(
            "Using Settlement Invoice ID is deprecated, see "
            "https://docs.mollie.com/reference/v2/settlements-api/get-settlement",
            APIDeprecationWarning,
        )
        return self._get_property("invoiceId")

    # Additional methods

    def is_open(self):
        return self._get_property("status") == self.STATUS_OPEN

    def is_pending(self):
        return self._get_property("status") == self.STATUS_PENDING

    def is_canceled(self):
        return self._get_property("status") == self.STATUS_PAIDOUT

    def is_failed(self):
        return self._get_property("status") == self.STATUS_FAILED

    @property
    def payments(self):
        """Return the payments related to this settlement."""
        return self.client.settlement_payments.on(self).list()

    @property
    def refunds(self):
        """Return the refunds related to this settlement."""
        return self.client.settlement_refunds.on(self).list()

    @property
    def chargebacks(self):
        """Return the chargebacks related to this settlement."""
        return self.client.settlement_chargebacks.on(self).list()



from .payments import Payments


class CustomerPayments(Payments):
    customer_id = None

    def get_resource_name(self):
        return f"customers/{self.customer_id}/payments"

    def with_parent_id(self, customer_id):
        self.customer_id = customer_id
        return self

    def on(self, customer):
        return self.with_parent_id(customer.id)



from .base import Base


class Payment(Base):
    @classmethod
    def get_resource_class(cls, client):
        from ..resources.payments import Payments

        return Payments(client)

    STATUS_OPEN = "open"
    STATUS_PENDING = "pending"
    STATUS_CANCELED = "canceled"
    STATUS_EXPIRED = "expired"
    STATUS_FAILED = "failed"
    STATUS_PAID = "paid"
    STATUS_AUTHORIZED = "authorized"

    SEQUENCETYPE_ONEOFF = "oneoff"
    SEQUENCETYPE_FIRST = "first"
    SEQUENCETYPE_RECURRING = "recurring"

    # Documented properties

    @property
    def resource(self):
        return self._get_property("resource")

    @property
    def id(self):
        return self._get_property("id")

    @property
    def mode(self):
        return self._get_property("mode")

    @property
    def created_at(self):
        return self._get_property("createdAt")

    @property
    def status(self):
        return self._get_property("status")

    @property
    def is_cancelable(self):
        return self._get_property("isCancelable")

    @property
    def authorized_at(self):
        return self._get_property("authorizedAt")

    @property
    def paid_at(self):
        return self._get_property("paidAt")

    @property
    def canceled_at(self):
        return self._get_property("canceledAt")

    @property
    def expires_at(self):
        return self._get_property("expiresAt")

    @property
    def expired_at(self):
        return self._get_property("expiredAt")

    @property
    def failed_at(self):
        return self._get_property("failedAt")

    @property
    def amount(self):
        return self._get_property("amount")

    @property
    def amount_refunded(self):
        return self._get_property("amountRefunded")

    @property
    def amount_remaining(self):
        return self._get_property("amountRemaining")

    @property
    def description(self):
        return self._get_property("description")

    @property
    def redirect_url(self):
        return self._get_property("redirectUrl")

    @property
    def webhook_url(self):
        return self._get_property("webhookUrl")

    @property
    def method(self):
        return self._get_property("method")

    @property
    def metadata(self):
        return self._get_property("metadata")

    @property
    def locale(self):
        return self._get_property("locale")

    @property
    def country_code(self):
        return self._get_property("countryCode")

    @property
    def profile_id(self):
        return self._get_property("profileId")

    @property
    def settlement_amount(self):
        return self._get_property("settlementAmount")

    @property
    def settlement_id(self):
        return self._get_property("settlementId")

    @property
    def customer_id(self):
        return self._get_property("customerId")

    @property
    def sequence_type(self):
        return self._get_property("sequenceType")

    @property
    def mandate_id(self):
        return self._get_property("mandateId")

    @property
    def subscription_id(self):
        return self._get_property("subscriptionId")

    @property
    def order_id(self):
        return self._get_property("orderId")

    @property
    def application_fee(self):
        return self._get_property("applicationFee")

    @property
    def details(self):
        return self._get_property("details")

    # documented _links

    @property
    def checkout_url(self):
        return self._get_link("checkout")

    @property
    def refunds(self):
        """Return the refunds related to this payment."""
        return self.client.payment_refunds.on(self).list()

    @property
    def chargebacks(self):
        """Return the chargebacks related to this payment."""
        return self.client.payment_chargebacks.on(self).list()

    @property
    def captures(self):
        """Return the captures related to this payment"""
        return self.client.captures.on(self).list()

    @property
    def settlement(self):
        """Return the settlement for this payment."""
        return self.client.settlements.get(self.settlement_id)

    @property
    def mandate(self):
        """Return the mandate for this payment."""
        return self.client.customer_mandates.with_parent_id(self.customer_id).get(self.mandate_id)

    @property
    def subscription(self):
        """Return the subscription for this payment."""
        return self.client.customer_subscriptions.with_parent_id(self.customer_id).get(self.subscription_id)

    @property
    def customer(self):
        """Return the customer for this payment."""
        return self.client.customers.get(self.customer_id)

    @property
    def order(self):
        """Return the order for this payment."""
        from ..resources.orders import Order

        url = self._get_link("order")
        if url:
            resp = self.client.orders.perform_api_call(self.client.orders.REST_READ, url)
            return Order(resp, self.client)

    # additional methods

    def is_open(self):
        return self._get_property("status") == self.STATUS_OPEN

    def is_pending(self):
        return self._get_property("status") == self.STATUS_PENDING

    def is_canceled(self):
        return self._get_property("status") == self.STATUS_CANCELED

    def is_expired(self):
        return self._get_property("status") == self.STATUS_EXPIRED

    def is_failed(self):
        return self._get_property("status") == self.STATUS_FAILED

    def is_authorized(self):
        return self._get_property("status") == self.STATUS_AUTHORIZED

    def is_paid(self):
        return self._get_property("paidAt") is not None

    def has_refunds(self):
        return self._get_link("refunds") is not None

    def can_be_refunded(self):
        return self._get_property("amountRemaining") is not None

    def has_sequence_type_first(self):
        return self._get_property("sequenceType") == self.SEQUENCETYPE_FIRST

    def has_sequence_type_recurring(self):
        return self._get_property("sequenceType") == self.SEQUENCETYPE_RECURRING



from .base import Base


class PaymentLink(Base):
    @classmethod
    def get_resource_class(cls, client):
        from ..resources.payment_links import PaymentLinks

        return PaymentLinks(client)

    @classmethod
    def get_object_name(cls):
        return "payment_links"

    # Documented properties

    @property
    def resource(self):
        return self._get_property("resource")

    @property
    def id(self):
        return self._get_property("id")

    @property
    def description(self):
        return self._get_property("description")

    @property
    def mode(self):
        return self._get_property("mode")

    @property
    def profile_id(self):
        return self._get_property("profileId")

    @property
    def amount(self):
        return self._get_property("amount")

    @property
    def redirect_url(self):
        return self._get_property("redirectUrl")

    @property
    def webhook_url(self):
        return self._get_property("webhookUrl")

    @property
    def created_at(self):
        return self._get_property("createdAt")

    @property
    def paid_at(self):
        return self._get_property("paidAt")

    @property
    def updated_at(self):
        return self._get_property("updatedAt")

    @property
    def expires_at(self):
        return self._get_property("expiresAt")

    # documented _links

    @property
    def payment_link(self):
        return self._get_link("paymentLink")

    # additional methods

    def is_paid(self):
        return self._get_property("paidAt") is not None

    def has_expiration_date(self):
        return self._get_property("expiresAt") is not None



from .base import Base


class Organization(Base):
    @classmethod
    def get_resource_class(cls, client):
        from ..resources.organizations import Organizations

        return Organizations(client)

    @property
    def id(self):
        return self._get_property("id")

    @property
    def name(self):
        return self._get_property("name")

    @property
    def email(self):
        return self._get_property("email")

    @property
    def locale(self):
        return self._get_property("locale")

    @property
    def address(self):
        return self._get_property("address")

    @property
    def registration_number(self):
        return self._get_property("registrationNumber")

    @property
    def vat_number(self):
        return self._get_property("vatNumber")

    @property
    def dashboard(self):
        return self._get_link("dashboard")



from ..error import DataConsistencyError
from ..objects.order_line import OrderLine
from .base import Base


class OrderLines(Base):
    order_id = None

    def get_resource_name(self):
        return f"orders/{self.order_id}/lines"

    def get_resource_object(self, result):
        return OrderLine(result, self.client)

    def with_parent_id(self, order_id):
        self.order_id = order_id
        return self

    def on(self, order):
        return self.with_parent_id(order.id)

    def delete(self, data, *args):
        """
        Custom handling for deleting orderlines.

        Orderlines are deleted by issuing a DELETE on the orders/*/lines endpoint,
        with the orderline IDs and quantities in the request body.
        """
        path = self.get_resource_name()
        result = self.perform_api_call(self.REST_DELETE, path, data=data)
        return result

    def update(self, resource_id, data=None, **params):
        """
        Custom handling for updating orderlines.

        The API returns an Order object. Since we are sending the request through an orderline object, it makes more
        sense to convert the returned object to to the updated orderline object.

        If you wish to retrieve the order object, you can do so by using the order_id property of the orderline.
        """
        path = self.get_resource_name() + "/" + str(resource_id)
        result = self.perform_api_call(self.REST_UPDATE, path, data=data)

        for line in result["lines"]:
            if line["id"] == resource_id:
                return self.get_resource_object(line)
        raise DataConsistencyError(f"Line id {resource_id} not found in response.")



from ..error import IdentifierError
from ..objects.payment_link import PaymentLink
from .base import Base


class PaymentLinks(Base):
    RESOURCE_ID_PREFIX = "pl_"

    def get_resource_object(self, result):
        return PaymentLink(result, self.client)

    def get_resource_name(self):
        return "payment-links"

    def get(self, payment_link_id, **params):
        if not payment_link_id or not payment_link_id.startswith(self.RESOURCE_ID_PREFIX):
            raise IdentifierError(
                f"Invalid payment ID: '{payment_link_id}'. A payment ID should start with '{self.RESOURCE_ID_PREFIX}'."
            )
        return super().get(payment_link_id, **params)



from .base import Base


class Mandate(Base):
    @classmethod
    def get_resource_class(cls, client):
        from ..resources.customer_mandates import CustomerMandates

        return CustomerMandates(client)

    STATUS_PENDING = "pending"
    STATUS_VALID = "valid"
    STATUS_INVALID = "invalid"

    @property
    def id(self):
        return self._get_property("id")

    @property
    def resource(self):
        return self._get_property("resource")

    @property
    def status(self):
        return self._get_property("status")

    @property
    def method(self):
        return self._get_property("method")

    @property
    def details(self):
        return self._get_property("details")

    @property
    def mandate_reference(self):
        return self._get_property("mandateReference")

    @property
    def signature_date(self):
        return self._get_property("signatureDate")

    @property
    def created_at(self):
        return self._get_property("createdAt")

    def is_pending(self):
        """Check if the mandate is pending."""
        return self.status == self.STATUS_PENDING

    def is_valid(self):
        """Check if the mandate is valid."""
        return self.status == self.STATUS_VALID

    def is_invalid(self):
        """Check if the mandate is invalid."""
        return self.status == self.STATUS_INVALID

    @property
    def customer(self):
        """Return the customer for this mandate."""
        from .customer import Customer  # work around circular imports

        url = self._get_link("customer")
        if url:
            resp = self.client.customers.perform_api_call(self.client.customers.REST_READ, url)
            return Customer(resp)



class Error(Exception):
    """Base exception."""

    pass


class RequestError(Error):
    """Errors while preparing or performing an API request."""

    pass


class RequestSetupError(RequestError):
    """Errors while preparing an API request."""

    pass


class IdentifierError(RequestSetupError):
    """Errors related to invalid resource identifiers that will be requested from the API."""

    pass


class ResponseHandlingError(Error):
    """Errors related to handling the response from the API."""

    pass


class ResponseError(Error):
    """Errors reported by the API."""

    status = None
    field = None

    def __init__(self, resp=None):
        message = resp["detail"]
        super().__init__(message)
        self.status = resp["status"]
        if "field" in resp:
            self.field = resp["field"]

    @staticmethod
    def factory(resp):
        """Return a ResponseError subclass based on the API payload.

        All errors are documented: https://docs.mollie.com/guides/handling-errors#all-possible-status-codes
        More exceptions should be added here when appropriate, and when useful examples of API errors are available.
        """
        status = resp["status"]
        if status == 401:
            return UnauthorizedError(resp)
        elif status == 404:
            return NotFoundError(resp)
        elif status == 422:
            return UnprocessableEntityError(resp)
        else:
            # generic fallback
            return ResponseError(resp)


class UnauthorizedError(ResponseError):
    """Your request wasn't executed due to failed authentication. Check your API key."""

    pass


class NotFoundError(ResponseError):
    """The object referenced by your API request does not exist."""

    pass


class UnprocessableEntityError(ResponseError):
    """We could not process your request due to another reason than the ones listed above.

    The response usually contains a field property to indicate which field is causing the issue.
    """

    pass


class DataConsistencyError(Error):
    """We could not process the API response due to an inconsistency.

    We received different data than expected from the API.
    """

    pass


class EmbedNotFound(Error):
    """We could not access requested data with an object because the required embed was missing.

    Some data within an object is only available after it is explicitly requested using an embed query parameter.
    You tried to access data which is available from an embed, but the data is missing: apparently you forgot
    to set the related embed parameter.
    """

    def __init__(self, embed_name):
        msg = (
            "You tried to access embedded data, but did not request to embed this data in the request. "
            f'Please specify embed="{embed_name}" when requesting the data.'
        )
        super().__init__(msg)


"""
Deprecation policy

When a minor version release will be done, remove all code that triggers the DeprecationWarning subclass.
Then rename the DeprecationWarning subclass to the next minor version (RemovedIn24Warning => RemovedIn25Warning).
and rename the PendingDeprecation subclass below also.

This will make all existing PendingDeprecationWarnings change to DeprecationWarnings.

This deprecation policy is similar to what is used in Django and Django-rest-framework.
See https://www.django-rest-framework.org/community/release-notes/#deprecation-policy for details.
"""


class RemovedIn27Warning(DeprecationWarning):
    """Deprecation warning for features that will be removed in version 2.7.0."""

    pass


class RemovedIn28Warning(PendingDeprecationWarning):
    """Pending deprecation warning for features that will be removed in version 2.8.0."""

    pass


class APIDeprecationWarning(DeprecationWarning):
    """
    Features that are deprecated in the Mollie API.

    This library cannot control when an API feature will be removed. We can only label it as 'deprecated'
    using this warning, and hope that library users stop using the feature before the removal.
    """

    pass



from .base import Base
from .customer import Customer


class Subscription(Base):
    @classmethod
    def get_resource_class(cls, client):
        from ..resources.customer_subscriptions import CustomerSubscriptions

        return CustomerSubscriptions(client)

    STATUS_ACTIVE = "active"
    STATUS_PENDING = "pending"  # Waiting for a valid mandate.
    STATUS_CANCELED = "canceled"
    STATUS_SUSPENDED = "suspended"  # Active, but mandate became invalid.
    STATUS_COMPLETED = "completed"

    @property
    def status(self):
        return self._get_property("status")

    def is_active(self):
        return self.status == self.STATUS_ACTIVE

    def is_pending(self):
        return self.status == self.STATUS_PENDING

    def is_canceled(self):
        return self.status == self.STATUS_CANCELED

    def is_suspended(self):
        return self.status == self.STATUS_SUSPENDED

    def is_completed(self):
        return self.status == self.STATUS_COMPLETED

    @property
    def resource(self):
        return self._get_property("resource")

    @property
    def id(self):
        return self._get_property("id")

    @property
    def mode(self):
        return self._get_property("mode")

    @property
    def created_at(self):
        return self._get_property("createdAt")

    @property
    def amount(self):
        return self._get_property("amount")

    @property
    def times(self):
        return int(self._get_property("times"))

    @property
    def interval(self):
        return self._get_property("interval")

    @property
    def start_date(self):
        return self._get_property("startDate")

    @property
    def description(self):
        return self._get_property("description")

    @property
    def method(self):
        return self._get_property("method")

    @property
    def canceled_at(self):
        return self._get_property("canceledAt")

    @property
    def webhook_url(self):
        return self._get_property("webhookUrl")

    @property
    def metadata(self):
        return self._get_property("metadata")

    @property
    def customer(self):
        """Return the customer for this subscription."""
        url = self._get_link("customer")
        if url:
            resp = self.client.customers.perform_api_call(self.client.customers.REST_READ, url)
            return Customer(resp)

    @property
    def payments(self):
        """Return a list of payments for this subscription."""
        payments = self.client.subscription_payments.on(self).list()
        return payments



from .base import Base


class Chargeback(Base):
    @classmethod
    def get_resource_class(cls, client):
        from ..resources.chargebacks import Chargebacks

        return Chargebacks(client)

    @property
    def id(self):
        return self._get_property("id")

    @property
    def amount(self):
        return self._get_property("amount")

    @property
    def settlement_amount(self):
        return self._get_property("settlementAmount")

    @property
    def created_at(self):
        return self._get_property("createdAt")

    @property
    def reversed_at(self):
        return self._get_property("reversedAt")

    @property
    def payment_id(self):
        return self._get_property("paymentId")



from .base import Base


class Capture(Base):
    @classmethod
    def get_resource_class(cls, client):
        from ..resources.captures import Captures

        return Captures(client)

    @property
    def id(self):
        return self._get_property("id")

    @property
    def mode(self):
        return self._get_property("mode")

    @property
    def amount(self):
        return self._get_property("amount")

    @property
    def settlement_amount(self):
        return self._get_property("settlementAmount")

    @property
    def payment_id(self):
        return self._get_property("paymentId")

    @property
    def shipment_id(self):
        return self._get_property("shipmentId")

    @property
    def settlement_id(self):
        return self._get_property("settlementId")

    @property
    def created_at(self):
        return self._get_property("createdAt")

    @property
    def payment(self):
        """Return the payment for this capture."""
        return self.client.payments.get(self.payment_id)

    @property
    def shipment(self):
        """Return the shipment for this capture."""
        from .shipment import Shipment

        url = self._get_link("shipment")

        if url:
            resp = self.client.orders.perform_api_call(self.client.orders.REST_READ, url)
            return Shipment(resp)

    @property
    def settlement(self):
        """Return the settlement for this capture."""
        return self.client.settlements.get(self.settlement_id)



import json
import platform
import re
import ssl
from collections import OrderedDict
from urllib.parse import urlencode

import requests
from requests_oauthlib import OAuth2Session
from urllib3.util import Retry

from .error import RequestError, RequestSetupError
from .resources.captures import Captures
from .resources.chargebacks import Chargebacks
from .resources.customer_mandates import CustomerMandates
from .resources.customer_payments import CustomerPayments
from .resources.customer_subscriptions import CustomerSubscriptions
from .resources.customers import Customers
from .resources.invoices import Invoices
from .resources.methods import Methods
from .resources.onboarding import Onboarding
from .resources.orders import Orders
from .resources.organizations import Organizations
from .resources.payment_chargebacks import PaymentChargebacks
from .resources.payment_links import PaymentLinks
from .resources.payment_refunds import PaymentRefunds
from .resources.payments import Payments
from .resources.permissions import Permissions
from .resources.profile_chargebacks import ProfileChargebacks
from .resources.profile_methods import ProfileMethods
from .resources.profile_payments import ProfilePayments
from .resources.profile_refunds import ProfileRefunds
from .resources.profiles import Profiles
from .resources.refunds import Refunds
from .resources.settlement_captures import SettlementCaptures
from .resources.settlement_chargebacks import SettlementChargebacks
from .resources.settlement_payments import SettlementPayments
from .resources.settlement_refunds import SettlementRefunds
from .resources.settlements import Settlements
from .resources.subscription_payments import SubscriptionPayments
from .resources.subscriptions import Subscriptions
from .version import VERSION


class Client(object):
    CLIENT_VERSION = VERSION
    API_ENDPOINT = "https://api.mollie.com"
    API_VERSION = "v2"
    UNAME = " ".join(platform.uname())

    OAUTH_AUTHORIZATION_URL = "https://www.mollie.com/oauth2/authorize"
    OAUTH_AUTO_REFRESH_URL = API_ENDPOINT + "/oauth2/tokens"
    OAUTH_TOKEN_URL = API_ENDPOINT + "/oauth2/tokens"

    @staticmethod
    def validate_api_endpoint(api_endpoint):
        return api_endpoint.strip().rstrip("/")

    @staticmethod
    def validate_api_key(api_key):
        api_key = api_key.strip()
        if not re.compile(r"^(live|test)_\w+$").match(api_key):
            raise RequestSetupError(f"Invalid API key: '{api_key}'. An API key must start with 'test_' or 'live_'.")
        return api_key

    @staticmethod
    def validate_access_token(access_token):
        access_token = access_token.strip()
        if not access_token.startswith("access_"):
            raise RequestSetupError(
                f"Invalid access token: '{access_token}'. An access token must start with 'access_'."
            )
        return access_token

    def __init__(self, api_endpoint=None, timeout=(2, 10), retry=3):
        """Initialize a new Mollie API client.

        :param api_endpoint: The API endpoint to communicate to, this default to the production environment (string)
        :param timeout: The timeouts used for the HTTP requests to the API, the default specifies both connect and
            read timeout (integer or tuple)
        :param retry: The number of retries that the client should perform in case of failed requests. Note that only
            connect errors trigger a retry, errors reponses from the API don't (integer).
        """
        self.api_endpoint = self.validate_api_endpoint(api_endpoint or self.API_ENDPOINT)
        self.api_version = self.API_VERSION
        self.timeout = timeout
        self.retry = retry
        self.api_key = None
        self._client = None

        self._oauth_client = None
        self.client_secret = None
        self.access_token = None
        self.set_token = None

        # add endpoint resources
        self.payments = Payments(self)
        self.payment_links = PaymentLinks(self)
        self.payment_refunds = PaymentRefunds(self)
        self.payment_chargebacks = PaymentChargebacks(self)
        self.profiles = Profiles(self)
        self.profile_chargebacks = ProfileChargebacks(self)
        self.profile_methods = ProfileMethods(self)
        self.profile_payments = ProfilePayments(self)
        self.profile_refunds = ProfileRefunds(self)
        self.methods = Methods(self)
        self.refunds = Refunds(self)
        self.chargebacks = Chargebacks(self)
        self.customers = Customers(self)
        self.customer_mandates = CustomerMandates(self)
        self.customer_subscriptions = CustomerSubscriptions(self)
        self.customer_payments = CustomerPayments(self)
        self.orders = Orders(self)
        self.organizations = Organizations(self)
        self.subscription_payments = SubscriptionPayments(self)
        self.invoices = Invoices(self)
        self.permissions = Permissions(self)
        self.onboarding = Onboarding(self)
        self.captures = Captures(self)
        self.settlements = Settlements(self)
        self.settlement_payments = SettlementPayments(self)
        self.settlement_refunds = SettlementRefunds(self)
        self.settlement_chargebacks = SettlementChargebacks(self)
        self.settlement_captures = SettlementCaptures(self)
        self.subscriptions = Subscriptions(self)

        # compose base user agent string
        self.user_agent_components = OrderedDict()
        self.set_user_agent_component("Mollie", self.CLIENT_VERSION)
        self.set_user_agent_component("Python", platform.python_version())
        self.set_user_agent_component(
            "OpenSSL", ssl.OPENSSL_VERSION.split(" ")[1], sanitize=False
        )  # keep legacy formatting of this component

    def set_api_endpoint(self, api_endpoint):
        self.api_endpoint = self.validate_api_endpoint(api_endpoint)

    def set_api_key(self, api_key):
        self.api_key = self.validate_api_key(api_key)

    def set_access_token(self, access_token):
        self.api_key = self.validate_access_token(access_token)
        self.set_user_agent_component("OAuth", "2.0", sanitize=False)  # keep spelling equal to the PHP client

    def set_timeout(self, timeout):
        self.timeout = timeout

    def set_user_agent_component(self, key, value, sanitize=True):
        """Add or replace new user-agent component strings.

        Given strings are formatted along the format agreed upon by Mollie and implementers:
        - key and values are separated by a forward slash ("/").
        - multiple key/values are separated by a space.
        - keys are camel-cased, and cannot contain spaces.
        - values cannot contain spaces.

        Note: When you set sanitize=false you need to make sure the formatting is correct yourself.
        """
        if sanitize:
            key = "".join(_x.capitalize() for _x in re.findall(r"\S+", key))
            if re.search(r"\s+", value):
                value = "_".join(re.findall(r"\S+", value))
        self.user_agent_components[key] = value

    @property
    def user_agent(self):
        """Return the formatted user agent string."""
        components = ["/".join(x) for x in self.user_agent_components.items()]
        return " ".join(components)

    def _format_request_data(self, path, data, params):
        if path.startswith(f"{self.api_endpoint}/{self.api_version}"):
            url = path
        else:
            url = f"{self.api_endpoint}/{self.api_version}/{path}"

        if data is not None:
            try:
                data = json.dumps(data)
            except Exception as err:
                raise RequestSetupError(f"Error encoding parameters into JSON: '{err}'.")

        querystring = generate_querystring(params)
        if querystring:
            url += "?" + querystring
            params = None

        return url, data, params

    def _perform_http_call_apikey(self, http_method, path, data=None, params=None):
        if not self.api_key:
            raise RequestSetupError("You have not set an API key. Please use set_api_key() to set the API key.")

        if not self._client:
            self._client = requests.Session()
            self._client.verify = True
            self._setup_retry()

        url, data, params = self._format_request_data(path, data, params)
        try:
            response = self._client.request(
                method=http_method,
                url=url,
                headers={
                    "Accept": "application/json",
                    "Authorization": f"Bearer {self.api_key}",
                    "Content-Type": "application/json",
                    "User-Agent": self.user_agent,
                    "X-Mollie-Client-Info": self.UNAME,
                },
                params=params,
                data=data,
                timeout=self.timeout,
            )
        except requests.exceptions.RequestException as err:
            raise RequestError(f"Unable to communicate with Mollie: {err}")
        return response

    def _perform_http_call_oauth(self, http_method, path, data=None, params=None):
        url, data, params = self._format_request_data(path, data, params)
        try:
            response = self._oauth_client.request(
                method=http_method,
                url=url,
                headers={
                    "Accept": "application/json",
                    "Content-Type": "application/json",
                    "User-Agent": self.user_agent,
                    "X-Mollie-Client-Info": self.UNAME,
                },
                params=params,
                data=data,
                timeout=self.timeout,
            )
        except requests.exceptions.RequestException as err:
            raise RequestError(f"Unable to communicate with Mollie: {err}")
        return response

    def perform_http_call(self, http_method, path, data=None, params=None):
        if self._oauth_client:
            return self._perform_http_call_oauth(http_method, path, data=data, params=params)
        else:
            return self._perform_http_call_apikey(http_method, path, data=data, params=params)

    def setup_oauth(self, client_id, client_secret, redirect_uri, scope, token, set_token):
        """
        :param client_id: (string)
        :param client_secret: (string)
        :param redirect_uri: (string)
        :param scope: Mollie connect permissions (list)
        :param token: The stored token (dict)
        :param set_token: Callable that stores a token (dict)
        :return: authorization url (url)
        """
        self.set_user_agent_component("OAuth", "2.0", sanitize=False)  # keep spelling equal to the PHP client
        self.set_token = set_token
        self.client_secret = client_secret
        self._oauth_client = OAuth2Session(
            client_id,
            auto_refresh_kwargs={
                "client_id": client_id,
                "client_secret": self.client_secret,
            },
            auto_refresh_url=self.OAUTH_AUTO_REFRESH_URL,
            redirect_uri=redirect_uri,
            scope=scope,
            token=token,
            token_updater=set_token,
        )
        self._oauth_client.verify = True
        self._setup_retry()

        authorization_url = None
        if not self._oauth_client.authorized:
            authorization_url, state = self._oauth_client.authorization_url(self.OAUTH_AUTHORIZATION_URL)

        # The merchant should visit this url to authorize access.
        return self._oauth_client.authorized, authorization_url

    def setup_oauth_authorization_response(self, authorization_response):
        """
        :param authorization_response: The full callback URL (string)
        :return: None
        """
        # Fetch an access token from the provider using the authorization code obtained during user authorization.
        self.access_token = self._oauth_client.fetch_token(
            self.OAUTH_TOKEN_URL,
            authorization_response=authorization_response,
            client_secret=self.client_secret,
        )
        self.set_token(self.access_token)
        return self.access_token

    def _setup_retry(self):
        """Configure a retry behaviour on the HTTP client."""
        if self.retry:
            retry = Retry(connect=self.retry, backoff_factor=1)
            adapter = requests.adapters.HTTPAdapter(max_retries=retry)

            if self._client:
                self._client.mount("https://", adapter)
            if self._oauth_client:
                self._oauth_client.mount("https://", adapter)


def generate_querystring(params):
    """
    Generate a querystring suitable for use in the v2 api.

    The Requests library doesn't know how to generate querystrings that encode dictionaries using square brackets:
    https://api.mollie.com/v2/methods?amount[value]=100.00&amount[currency]=USD

    Note: we use `sorted()` to work around a difference in iteration behaviour between Python 2 and 3.
    This makes the output predictable, and ordering of querystring parameters shouldn't matter.
    """
    if not params:
        return None
    parts = []
    for param, value in sorted(params.items()):
        if not isinstance(value, dict):
            parts.append(urlencode({param: value}))
        else:
            # encode dictionary with square brackets
            for key, sub_value in sorted(value.items()):
                composed = f"{param}[{key}]"
                parts.append(urlencode({composed: sub_value}))
    if parts:
        return "&".join(parts)



from ..error import IdentifierError
from ..objects.customer import Customer
from .base import Base


class Customers(Base):
    RESOURCE_ID_PREFIX = "cst_"

    def get_resource_object(self, result):
        return Customer(result, self.client)

    def get(self, customer_id, **params):
        if not customer_id or not customer_id.startswith(self.RESOURCE_ID_PREFIX):
            raise IdentifierError(
                f"Invalid customer ID: '{customer_id}'. A customer ID should start with '{self.RESOURCE_ID_PREFIX}'."
            )
        return super().get(customer_id, **params)



from .base import Base


class Invoice(Base):
    @classmethod
    def get_resource_class(cls, client):
        from ..resources.invoices import Invoices

        return Invoices(client)

    @property
    def id(self):
        return self._get_property("id")

    @property
    def reference(self):
        return self._get_property("reference")

    @property
    def vat_number(self):
        return self._get_property("vatNumber")

    @property
    def status(self):
        return self._get_property("status")

    @property
    def issued_at(self):
        return self._get_property("issuedAt")

    @property
    def due_at(self):
        return self._get_property("dueAt")

    @property
    def resource(self):
        return self._get_property("resource")

    @property
    def net_amount(self):
        return self._get_property("netAmount")

    @property
    def vat_amount(self):
        return self._get_property("vatAmount")

    @property
    def gross_amount(self):
        return self._get_property("grossAmount")

    @property
    def lines(self):
        return self._get_property("lines") or []

    @property
    def pdf(self):
        return self._get_link("pdf")



from ..error import IdentifierError
from ..objects.order import Order
from .base import Base


class Orders(Base):
    RESOURCE_ID_PREFIX = "ord_"

    def get_resource_object(self, result):
        return Order(result, self.client)

    def get(self, order_id, **params):
        if not order_id or not order_id.startswith(self.RESOURCE_ID_PREFIX):
            raise IdentifierError(
                f"Invalid order ID: '{order_id}'. An order ID should start with '{self.RESOURCE_ID_PREFIX}'."
            )

        result_order = super().get(order_id, **params)

        requested_embeds = self.extract_embed(params)
        if requested_embeds:
            result_order.requested_embeds = requested_embeds

        return result_order

    def delete(self, order_id, data=None):
        """Cancel order and return the order object.

        Deleting an order causes the order status to change to canceled.
        The updated order object is returned.
        """
        if not order_id or not order_id.startswith(self.RESOURCE_ID_PREFIX):
            raise IdentifierError(
                f"Invalid order ID: '{order_id}'. An order ID should start with '{self.RESOURCE_ID_PREFIX}'."
            )
        result = super().delete(order_id, data)
        return self.get_resource_object(result)



from .payments import Payments


class SubscriptionPayments(Payments):
    customer_id = None
    subscription_id = None

    def get_resource_name(self):
        return f"customers/{self.customer_id}/subscriptions/{self.subscription_id}/payments"

    def with_parent_id(self, customer_id, subscription_id):
        self.customer_id = customer_id
        self.subscription_id = subscription_id
        return self

    def on(self, subscription):
        # TODO: A request has been filed to mollie to add the customer id to the subscription response.
        return self.with_parent_id(subscription.customer.id, subscription.id)



# In this file the version of the package is defined.

# Don't change the syntax of the definition unless you know what you're doing, because this file is
# processed by python imports and by regular expressions. The version is defined as a string in the
# regular semantic versioning scheme (major,minor,patch).

VERSION = "2.9.1"



from .payments import Payments


class OrderPayments(Payments):
    order_id = None

    def get_resource_name(self):
        return f"orders/{self.order_id}/payments"

    def with_parent_id(self, order_id):
        self.order_id = order_id
        return self

    def on(self, order):
        return self.with_parent_id(order.id)



from ..objects.subscription import Subscription
from .base import Base


class Subscriptions(Base):
    def get_resource_object(self, result):
        return Subscription(result, self.client)

    def create(self, data=None, **params):
        raise NotImplementedError('The endpoint "create" is not supported.')

    def get(self, resource_id, **params):
        raise NotImplementedError('The endpoint "get" is not supported.')

    def update(self, resource_id, data=None, **params):
        raise NotImplementedError('The endpoint "update" is not supported.')

    def delete(self, resource_id, data=None):
        raise NotImplementedError('The endpoint "delete" is not supported.')



from .base import Base


class List(Base):
    current = None

    def __init__(self, result, object_type, client=None):
        super().__init__(result, client)
        self.object_type = object_type

    def __len__(self):
        """Return the count field."""
        return int(self["count"])

    def __iter__(self):
        """Implement iterator interface."""
        self.current = None
        return self

    def __next__(self):
        """Implement iterator interface."""
        if self.current is None:
            self.current = 0
        else:
            self.current += 1
        try:
            item = self["_embedded"][self.object_type.get_object_name()][self.current]
            return self.object_type(item, self.client)
        except IndexError:
            self.current = None
            raise StopIteration

    next = __next__  # support python2 iterator interface

    @property
    def count(self):
        if "count" not in self:
            return None
        return int(self["count"])

    def has_next(self):
        """Return True if the list contains an url for the next set"""
        return self._get_link("next") is not None

    def has_previous(self):
        """Return True if the list contains an url for the previous set"""
        return self._get_link("previous") is not None

    def get_next(self):
        """Return the next set of objects in a list"""
        url = self._get_link("next")
        resource = self.object_type.get_resource_class(self.client)
        resp = resource.perform_api_call(resource.REST_READ, url)
        return List(resp, self.object_type, self.client)

    def get_previous(self):
        """Return the previous set of objects in a list"""
        url = self._get_link("previous")
        resource = self.object_type.get_resource_class(self.client)
        resp = resource.perform_api_call(resource.REST_READ, url)
        return List(resp, self.object_type, self.client)



class Base(dict):
    def __init__(self, data, client=None):
        """
        Create a new object from API result data.
        """
        super().__init__(data)
        self.client = client

    def _get_property(self, name):
        """Return the named property from dictionary values."""
        if name not in self:
            return None
        return self[name]

    def _get_link(self, name):
        """Return a link by its name."""
        try:
            return self["_links"][name]["href"]
        except (KeyError, TypeError):
            return None

    @classmethod
    def get_object_name(cls):
        name = cls.__name__.lower()
        return f"{name}s"

    @classmethod
    def get_resource_class(cls, client):
        raise NotImplementedError



from .base import Base


class OrderLine(Base):
    @classmethod
    def get_resource_class(cls, client):
        from ..resources.order_lines import OrderLines

        return OrderLines(client)

    STATUS_CREATED = "created"
    STATUS_AUTHORIZED = "authorized"
    STATUS_PAID = "paid"
    STATUS_SHIPPING = "shipping"
    STATUS_CANCELED = "canceled"
    STATUS_COMPLETED = "completed"

    @classmethod
    def get_object_name(cls):
        return "lines"

    @property
    def id(self):
        return self._get_property("id")

    @property
    def resource(self):
        return self._get_property("resource")

    @property
    def order_id(self):
        return self._get_property("orderId")

    @property
    def type(self):
        return self._get_property("type")

    @property
    def name(self):
        return self._get_property("name")

    @property
    def status(self):
        return self._get_property("status")

    @property
    def is_cancelable(self):
        return self._get_property("isCancelable")

    @property
    def quantity(self):
        return self._get_property("quantity")

    @property
    def quantity_shipped(self):
        return self._get_property("quantityShipped")

    @property
    def amount_shipped(self):
        return self._get_property("amountShipped")

    @property
    def quantity_refunded(self):
        return self._get_property("quantityRefunded")

    @property
    def amount_refunded(self):
        return self._get_property("amountRefunded")

    @property
    def quantity_canceled(self):
        return self._get_property("quantityCanceled")

    @property
    def amount_canceled(self):
        return self._get_property("amountCanceled")

    @property
    def shippable_quantity(self):
        return self._get_property("shippableQuantity")

    @property
    def refundable_quantity(self):
        return self._get_property("refundableQuantity")

    @property
    def cancelable_quantity(self):
        return self._get_property("cancelableQuantity")

    @property
    def unit_price(self):
        return self._get_property("unitPrice")

    @property
    def discount_amount(self):
        return self._get_property("discountAmount")

    @property
    def total_amount(self):
        return self._get_property("totalAmount")

    @property
    def vat_rate(self):
        return self._get_property("vatRate")

    @property
    def vat_amount(self):
        return self._get_property("vatAmount")

    @property
    def sku(self):
        return self._get_property("sku")

    @property
    def image_url(self):
        return self._get_link("imageUrl")

    @property
    def product_url(self):
        return self._get_link("productUrl")

    @property
    def created_at(self):
        return self._get_property("createdAt")

    @property
    def metadata(self):
        return self._get_property("metadata")

    # additional methods

    def is_created(self):
        return self._get_property("status") == self.STATUS_CREATED

    def is_authorized(self):
        return self._get_property("status") == self.STATUS_AUTHORIZED

    def is_paid(self):
        return self._get_property("status") == self.STATUS_PAID

    def is_shipping(self):
        return self._get_property("status") == self.STATUS_SHIPPING

    def is_canceled(self):
        return self._get_property("status") == self.STATUS_CANCELED

    def is_completed(self):
        return self._get_property("status") == self.STATUS_COMPLETED



from ..error import ResponseError, ResponseHandlingError
from ..objects.list import List


class Base(object):
    REST_CREATE = "POST"
    REST_UPDATE = "PATCH"
    REST_READ = "GET"
    REST_LIST = "GET"
    REST_DELETE = "DELETE"
    DEFAULT_LIMIT = 10

    def __init__(self, client):
        self.client = client

    def get_resource_object(self, result):
        raise NotImplementedError()

    def get_resource_name(self):
        return self.__class__.__name__.lower()

    def create(self, data=None, **params):
        path = self.get_resource_name()
        result = self.perform_api_call(self.REST_CREATE, path, data, params)
        return self.get_resource_object(result)

    def get(self, resource_id, **params):
        path = self.get_resource_name() + "/" + str(resource_id)
        result = self.perform_api_call(self.REST_READ, path, params=params)
        return self.get_resource_object(result)

    def update(self, resource_id, data=None, **params):
        path = self.get_resource_name() + "/" + str(resource_id)
        result = self.perform_api_call(self.REST_UPDATE, path, data, params)
        return self.get_resource_object(result)

    def delete(self, resource_id, data=None):
        path = self.get_resource_name() + "/" + str(resource_id)
        return self.perform_api_call(self.REST_DELETE, path, data)

    def list(self, **params):
        path = self.get_resource_name()
        result = self.perform_api_call(self.REST_LIST, path, params=params)
        return List(result, self.get_resource_object({}).__class__, self.client)

    def perform_api_call(self, http_method, path, data=None, params=None):
        resp = self.client.perform_http_call(http_method, path, data, params)
        if "application/hal+json" in resp.headers.get("Content-Type", ""):
            # set the content type according to the media type definition
            resp.encoding = "utf-8"
        try:
            result = resp.json() if resp.status_code != 204 else {}
        except Exception:
            raise ResponseHandlingError(
                f"Unable to decode Mollie API response (status code: {resp.status_code}): '{resp.text}'."
            )
        if resp.status_code < 200 or resp.status_code > 299:
            if "status" in result and (result["status"] < 200 or result["status"] > 299):
                # the factory will return the appropriate ResponseError subclass based on the result
                raise ResponseError.factory(result)
            else:
                raise ResponseHandlingError(
                    "Received HTTP error from Mollie API, but no status in payload "
                    f"(status code: {resp.status_code}): '{resp.text}'."
                )
        return result

    @staticmethod
    def extract_embed(params):
        """Extract and parse the embed parameter from the request."""
        if "embed" not in params:
            return
        return params["embed"].split(",")



from .base import Base
from .list import List
from .order_line import OrderLine


class Shipment(Base):
    @property
    def resource(self):
        return self._get_property("resource")

    @property
    def id(self):
        return self._get_property("id")

    @property
    def order_id(self):
        return self._get_property("orderId")

    @property
    def created_at(self):
        return self._get_property("createdAt")

    @property
    def tracking(self):
        return self._get_property("tracking")

    @property
    def tracking_url(self):
        return self.tracking["url"] if self.has_tracking_url() else None

    @property
    def lines(self):
        """Return the order lines of this shipment."""
        lines = self._get_property("lines") or []
        result = {
            "_embedded": {
                "lines": lines,
            },
            "count": len(lines),
        }
        return List(result, OrderLine, self.client)

    @property
    def order(self):
        """Return the order of this shipment."""
        return self.client.orders.get(self.order_id)

    # additional methods

    def has_tracking(self):
        return self.tracking is not None

    def has_tracking_url(self):
        return self.has_tracking() and self.tracking["url"] is not None



from ..error import IdentifierError
from ..objects.capture import Capture
from .base import Base


class Captures(Base):
    RESOURCE_ID_PREFIX = "cpt_"

    def get_resource_object(self, result):
        return Capture(result, self.client)

    def get_resource_name(self):
        return f"payments/{self.payment_id}/captures"

    def get(self, capture_id, **params):
        """Verify the capture ID and retrieve the capture from the API."""
        if not capture_id or not capture_id.startswith(self.RESOURCE_ID_PREFIX):
            raise IdentifierError(
                f"Invalid capture ID: '{capture_id}'. A capture ID should start with '{self.RESOURCE_ID_PREFIX}'."
            )
        return super().get(capture_id, **params)

    def with_parent_id(self, payment_id):
        self.payment_id = payment_id
        return self

    def on(self, payment):
        return self.with_parent_id(payment.id)



import re

from ..error import IdentifierError
from ..objects.settlement import Settlement
from .base import Base


class Settlements(Base):
    RESOURCE_ID_PREFIX = "stl_"

    # According to Mollie, the bank reference is formatted as:
    # - The Mollie customer ID, 4 to 7 digits.
    # - The year and month, 4 digits
    # - The sequence number of the settlement in that month, 2 digits
    # The components are separated by a dot.
    BANK_REFERENCE_REGEX = re.compile(r"^\d{4,7}\.\d{4}\.\d{2}$", re.ASCII)

    def get_resource_object(self, result):
        return Settlement(result, self.client)

    @classmethod
    def validate_settlement_id(cls, settlement_id):
        if not settlement_id or (
            not settlement_id.startswith(cls.RESOURCE_ID_PREFIX)
            and settlement_id not in ["next", "open"]
            and not cls.BANK_REFERENCE_REGEX.match(settlement_id)
        ):
            raise IdentifierError(
                f"Invalid settlement ID: '{settlement_id}'. A settlement ID should "
                f"start with '{cls.RESOURCE_ID_PREFIX}' "
                ", be 'next' or 'open' or contain a valid bank reference."
            )

    def get(self, settlement_id, **params):
        self.validate_settlement_id(settlement_id)
        """Verify the settlement ID and retrieve the settlement from the API."""
        return super().get(settlement_id, **params)



from ..error import IdentifierError
from ..objects.refund import Refund
from .base import Base


class Refunds(Base):
    RESOURCE_ID_PREFIX = "re_"

    def get_resource_object(self, result):
        return Refund(result, self.client)

    def get(self, refund_id, **params):
        if not refund_id or not refund_id.startswith(self.RESOURCE_ID_PREFIX):
            raise IdentifierError(
                f"Invalid refund ID: '{refund_id}'. A refund ID should start with '{self.RESOURCE_ID_PREFIX}'."
            )
        return super().get(refund_id, **params)



import re

from ..error import IdentifierError
from ..objects.permission import Permission
from .base import Base


class Permissions(Base):
    def get_resource_object(self, result):
        return Permission(result, self.client)

    @staticmethod
    def validate_permission_id(permission_id):
        if not permission_id or not bool(re.match(r"^[a-z]+\.[a-z]+$", permission_id)):
            raise IdentifierError(f"Invalid permission ID: '{permission_id}'. Does not match ^[a-z]+.[a-z]+$")

    def get(self, permission_id, **params):
        self.validate_permission_id(permission_id)
        return super().get(permission_id, **params)



from ..error import IdentifierError
from ..objects.mandate import Mandate
from .base import Base


class CustomerMandates(Base):
    RESOURCE_ID_PREFIX = "mdt_"
    customer_id = None

    def get_resource_object(self, result):
        return Mandate(result, self.client)

    def get(self, mandate_id, **params):
        if not mandate_id or not mandate_id.startswith(self.RESOURCE_ID_PREFIX):
            raise IdentifierError(
                f"Invalid mandate ID: '{mandate_id}'. A mandate ID should start with '{self.RESOURCE_ID_PREFIX}'."
            )
        return super().get(mandate_id, **params)

    def get_resource_name(self):
        return f"customers/{self.customer_id}/mandates"

    def with_parent_id(self, customer_id):
        self.customer_id = customer_id
        return self

    def on(self, customer):
        return self.with_parent_id(customer.id)



from .base import Base


class Issuer(Base):
    @property
    def resource(self):
        return self._get_property("resource")

    @property
    def id(self):
        return self._get_property("id")

    @property
    def name(self):
        return self._get_property("name")

    @property
    def image_svg(self):
        try:
            images = self._get_property("image")
            return images["svg"]
        except KeyError:
            return None

    @property
    def image_size1x(self):
        try:
            images = self._get_property("image")
            return images["size1x"]
        except KeyError:
            return None

    @property
    def image_size2x(self):
        try:
            images = self._get_property("image")
            return images["size2x"]
        except KeyError:
            return None



from .chargebacks import Chargebacks


class SettlementChargebacks(Chargebacks):
    settlement_id = None

    def get_resource_name(self):
        return f"settlements/{self.settlement_id}/chargebacks"

    def with_parent_id(self, settlement_id):
        self.settlement_id = settlement_id
        return self

    def on(self, settlement):
        return self.with_parent_id(settlement.id)



from ..error import IdentifierError
from ..objects.invoice import Invoice
from .base import Base


class Invoices(Base):
    RESOURCE_ID_PREFIX = "inv_"

    def get_resource_object(self, result):
        return Invoice(result, self.client)

    def get(self, invoice_id, **params):
        if not invoice_id or not invoice_id.startswith(self.RESOURCE_ID_PREFIX):
            raise IdentifierError(
                f"Invalid invoice ID: '{invoice_id}'. An invoice ID should start with '{self.RESOURCE_ID_PREFIX}'."
            )
        return super().get(invoice_id, **params)



from .base import Base


class Profile(Base):
    STATUS_UNVERIFIED = "unverified"
    STATUS_VERIFIED = "verified"
    STATUS_BLOCKED = "blocked"

    @classmethod
    def get_resource_class(cls, client):
        from ..resources.profiles import Profiles

        return Profiles(client)

    @property
    def id(self):
        return self._get_property("id")

    @property
    def resource(self):
        return self._get_property("resource")

    @property
    def mode(self):
        return self._get_property("mode")

    @property
    def name(self):
        return self._get_property("name")

    @property
    def website(self):
        return self._get_property("website")

    @property
    def email(self):
        return self._get_property("email")

    @property
    def phone(self):
        return self._get_property("phone")

    @property
    def category_code(self):
        return self._get_property("categoryCode")

    @property
    def status(self):
        return self._get_property("status")

    @property
    def review(self):
        return self._get_property("review")

    @property
    def created_at(self):
        return self._get_property("createdAt")

    @property
    def chargebacks(self):
        return self.client.profile_chargebacks.on(self).list()

    @property
    def methods(self):
        return self.client.profile_methods.on(self).list()

    @property
    def payments(self):
        return self.client.profile_payments.on(self).list()

    @property
    def refunds(self):
        return self.client.profile_refunds.on(self).list()

    @property
    def checkout_preview_url(self):
        return self._get_link("checkoutPreviewUrl")

    # additional methods

    def is_unverified(self):
        return self._get_property("status") == self.STATUS_UNVERIFIED

    def is_verified(self):
        return self._get_property("status") == self.STATUS_VERIFIED

    def is_blocked(self):
        return self._get_property("status") == self.STATUS_BLOCKED



from ..error import EmbedNotFound
from ..resources.order_lines import OrderLines
from ..resources.order_payments import OrderPayments
from ..resources.order_refunds import OrderRefunds
from ..resources.shipments import Shipments
from .base import Base
from .list import List
from .order_line import OrderLine
from .payment import Payment


class Order(Base):
    requested_embeds = None

    def __init__(self, data, client=None, requested_embeds=None):
        super().__init__(data, client)
        self.requested_embeds = requested_embeds

    def _has_embed(self, embed_name):
        if self.requested_embeds and embed_name in self.requested_embeds:
            return True
        return False

    @classmethod
    def get_resource_class(cls, client):
        from ..resources.orders import Orders

        return Orders(client)

    STATUS_CREATED = "created"
    STATUS_PAID = "paid"
    STATUS_AUTHORIZED = "authorized"
    STATUS_CANCELED = "canceled"
    STATUS_SHIPPING = "shipping"
    STATUS_COMPLETED = "completed"
    STATUS_EXPIRED = "expired"

    @property
    def id(self):
        return self._get_property("id")

    @property
    def resource(self):
        return self._get_property("resource")

    @property
    def profile_id(self):
        return self._get_property("profileId")

    @property
    def method(self):
        return self._get_property("method")

    @property
    def mode(self):
        return self._get_property("mode")

    @property
    def amount(self):
        return self._get_property("amount")

    @property
    def amount_captured(self):
        return self._get_property("amountCaptured")

    @property
    def amount_refunded(self):
        return self._get_property("amountRefunded")

    @property
    def status(self):
        return self._get_property("status")

    @property
    def is_cancelable(self):
        return self._get_property("isCancelable")

    @property
    def billing_address(self):
        return self._get_property("billingAddress")

    @property
    def consumer_date_of_birth(self):
        return self._get_property("consumerDateOfBirth")

    @property
    def order_number(self):
        return self._get_property("orderNumber")

    @property
    def shipping_address(self):
        return self._get_property("shippingAddress")

    @property
    def locale(self):
        return self._get_property("locale")

    @property
    def metadata(self):
        return self._get_property("metadata")

    @property
    def redirect_url(self):
        return self._get_property("redirectUrl")

    @property
    def webhook_url(self):
        return self._get_property("webhookUrl")

    @property
    def created_at(self):
        return self._get_property("createdAt")

    @property
    def expires_at(self):
        return self._get_property("expiresAt")

    @property
    def expired_at(self):
        return self._get_property("expiredAt")

    @property
    def paid_at(self):
        return self._get_property("paidAt")

    @property
    def authorized_at(self):
        return self._get_property("authorizedAt")

    @property
    def canceled_at(self):
        return self._get_property("canceledAt")

    @property
    def completed_at(self):
        return self._get_property("completedAt")

    # documented _links

    @property
    def checkout_url(self):
        return self._get_link("checkout")

    # additional methods

    def is_created(self):
        return self._get_property("status") == self.STATUS_CREATED

    def is_paid(self):
        return self._get_property("status") == self.STATUS_PAID

    def is_authorized(self):
        return self._get_property("status") == self.STATUS_AUTHORIZED

    def is_canceled(self):
        return self._get_property("status") == self.STATUS_CANCELED

    def is_shipping(self):
        return self._get_property("status") == self.STATUS_SHIPPING

    def is_completed(self):
        return self._get_property("status") == self.STATUS_COMPLETED

    def is_expired(self):
        return self._get_property("status") == self.STATUS_EXPIRED

    def create_refund(self, data=None, **params):
        """Create a refund for the order. When no data arg is given, a refund for all order lines is assumed."""
        if data is None:
            data = {"lines": []}
        refund = OrderRefunds(self.client).on(self).create(data, **params)
        return refund

    def cancel_lines(self, data=None):
        """Cancel the lines given. When no lines are given, cancel all the lines.

        Canceling an order line causes the order line status to change to canceled.
        An empty dictionary will be returned.
        """
        from ..resources.order_lines import OrderLines

        if data is None:
            data = {"lines": []}
        canceled = OrderLines(self.client).on(self).delete(data)
        return canceled

    @property
    def refunds(self):
        refunds = OrderRefunds(self.client).on(self).list()
        return refunds

    @property
    def lines(self):
        lines = self._get_property("lines") or []
        result = {
            "_embedded": {
                "lines": lines,
            },
            "count": len(lines),
        }
        return List(result, OrderLine, self.client)

    def update_line(self, resource_id, data):
        """Update a line for an order."""
        return OrderLines(self.client).on(self).update(resource_id, data)

    @property
    def shipments(self):
        """Retrieve all shipments for an order."""
        return Shipments(self.client).on(self).list()

    def create_shipment(self, data=None):
        """Create a shipment for an order. When no data arg is given, a shipment for all order lines is assumed."""
        if data is None:
            data = {"lines": []}
        return Shipments(self.client).on(self).create(data)

    def get_shipment(self, resource_id):
        """Retrieve a single shipment by a shipment's ID."""
        return Shipments(self.client).on(self).get(resource_id)

    def update_shipment(self, resource_id, data):
        """Update the tracking information of a shipment."""
        return Shipments(self.client).on(self).update(resource_id, data)

    @property
    def payments(self):
        if not self._has_embed("payments"):
            raise EmbedNotFound("payments")

        try:
            payments = self["_embedded"]["payments"]
        except KeyError:
            # No data available at API
            payments = []

        result = {
            "_embedded": {
                "payments": payments,
            },
            "count": len(payments),
        }
        return List(result, Payment, self.client)

    def create_payment(self, data):
        """Creates a new payment object for an order."""
        return OrderPayments(self.client).on(self).create(data)



from .base import Base
from .list import List
from .order_line import OrderLine


class Refund(Base):
    @classmethod
    def get_resource_class(cls, client):
        from ..resources.refunds import Refunds

        return Refunds(client)

    STATUS_QUEUED = "queued"
    STATUS_PENDING = "pending"
    STATUS_PROCESSING = "processing"
    STATUS_REFUNDED = "refunded"

    # documented properties

    @property
    def resource(self):
        return self._get_property("resource")

    @property
    def id(self):
        return self._get_property("id")

    @property
    def amount(self):
        return self._get_property("amount")

    @property
    def settlement_amount(self):
        return self._get_property("settlementAmount")

    @property
    def description(self):
        return self._get_property("description")

    @property
    def status(self):
        return self._get_property("status")

    @property
    def lines(self):
        """Return the lines for this refund."""
        lines = self._get_property("lines") or []
        result = {
            "_embedded": {
                "lines": lines,
            },
            "count": len(lines),
        }
        return List(result, OrderLine, self.client)

    @property
    def payment_id(self):
        return self._get_property("paymentId")

    @property
    def order_id(self):
        return self._get_property("orderId")

    @property
    def created_at(self):
        return self._get_property("createdAt")

    @property
    def metadata(self):
        return self._get_property("metadata")

    # documented _links

    @property
    def payment(self):
        """Return the payment for this refund."""
        return self.client.payments.get(self.payment_id)

    # @property
    # def settlement(self):
    #     """
    #     Return the settlement for this refund.

    #     TODO: Before we can return an Settlement object, we need to implement the Setlement API.
    #     """
    #     pass

    @property
    def order(self):
        """Return the order for this refund."""
        return self.client.orders.get(self.order_id)

    # additional methods

    def is_queued(self):
        return self.status == self.STATUS_QUEUED

    def is_pending(self):
        return self.status == self.STATUS_PENDING

    def is_processing(self):
        return self.status == self.STATUS_PROCESSING

    def is_refunded(self):
        return self.status == self.STATUS_REFUNDED



from ..error import IdentifierError
from ..objects.profile import Profile
from .base import Base


class Profiles(Base):
    RESOURCE_ID_PREFIX = "pfl_"

    def get_resource_object(self, result):
        return Profile(result, self.client)

    def get(self, profile_id, **params):
        if not profile_id or (not profile_id.startswith(self.RESOURCE_ID_PREFIX) and not profile_id == "me"):
            raise IdentifierError(
                f"Invalid profile ID: '{profile_id}'. A profile ID should start with '{self.RESOURCE_ID_PREFIX}' "
                "or it should be 'me'."
            )
        return super().get(profile_id, **params)



from .chargebacks import Chargebacks


class PaymentChargebacks(Chargebacks):
    payment_id = None

    def get_resource_name(self):
        return f"payments/{self.payment_id}/chargebacks"

    def with_parent_id(self, payment_id):
        self.payment_id = payment_id
        return self

    def on(self, payment):
        return self.with_parent_id(payment.id)



from .base import Base


class Permission(Base):
    @classmethod
    def get_resource_class(cls, client):
        from ..resources.permissions import Permissions

        return Permissions(client)

    @property
    def id(self):
        return self._get_property("id")

    @property
    def resource(self):
        return self._get_property("resource")

    @property
    def description(self):
        return self._get_property("description")

    @property
    def granted(self):
        return self._get_property("granted")



from ..error import IdentifierError
from ..objects.onboarding import Onboarding as OnboardingObject
from .base import Base


class Onboarding(Base):
    def get_resource_object(self, result):
        return OnboardingObject(result, self.client)

    def get(self, onboarding_id, **params):
        if not onboarding_id or not onboarding_id == "me":
            raise IdentifierError(f"Invalid onboarding ID: '{onboarding_id}'. An onboarding ID should be 'me'.")
        return super().get(onboarding_id, **params)

    def create(self, resource_id, data=None, **params):
        path = self.get_resource_name() + "/" + str(resource_id)
        result = self.perform_api_call(self.REST_CREATE, path, data, params)
        return self.get_resource_object(result)



import os.path
import re

from setuptools import find_packages, setup

ROOT_DIR = os.path.abspath(os.path.dirname(__file__))


def get_long_description():
    return open(os.path.join(ROOT_DIR, "README.md"), encoding="utf-8").read()


def get_version():
    """Read the version from a file (mollie/api/version.py) in the repository.

    We can't import here since we might import from an installed version.
    """
    version_file = open(os.path.join(ROOT_DIR, "mollie", "api", "version.py"), encoding="utf=8")
    contents = version_file.read()
    match = re.search(r'VERSION = [\'"]([^\'"]+)', contents)
    if match:
        return match.group(1)
    else:
        raise RuntimeError("Can't determine package version")


setup(
    name="mollie-api-python",
    version=get_version(),
    license="BSD",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    description="Mollie API client for Python",
    author="Mollie B.V.",
    author_email="info@mollie.com",
    maintainer="Four Digits B.V.",
    maintainer_email="info@fourdigits.nl",
    keywords=[
        "mollie",
        "payment",
        "service",
        "ideal",
        "creditcard",
        "mistercash",
        "bancontact",
        "sofort",
        "sofortbanking",
        "sepa",
        "paypal",
        "paysafecard",
        "podiumcadeaukaart",
        "banktransfer",
        "direct debit",
        "belfius",
        "belfius direct net",
        "kbc",
        "cbc",
        "refunds",
        "payments",
        "gateway",
        "gift cards",
        "intersolve",
        "fashioncheque",
        "podium cadeaukaart",
        "yourgift",
        "vvv giftcard",
        "webshop giftcard",
        "nationale entertainment card",
        "ing homepay",
        "klarna pay later",
        "klarna slice it",
        "przelewy24",
    ],
    url="https://github.com/mollie/mollie-api-python",
    install_requires=[
        "requests",
        "urllib3",
        "requests_oauthlib",
    ],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Topic :: Office/Business :: Financial",
    ],
)



from .base import Base


class Onboarding(Base):
    @classmethod
    def get_resource_class(cls, client):
        from ..resources.onboarding import Onboarding as OnboardingResource

        return OnboardingResource(client)

    STATUS_NEEDS_DATA = "needs-data"
    STATUS_IN_REVIEW = "in-review"  # Waiting for a valid mandate.
    STATUS_COMPLETED = "completed"

    @property
    def resource(self):
        return self._get_property("resource")

    @property
    def name(self):
        return self._get_property("name")

    @property
    def signed_up_at(self):
        return self._get_property("signedUpAt")

    @property
    def status(self):
        return self._get_property("status")

    @property
    def can_receive_payments(self):
        return self._get_property("canReceivePayments")

    @property
    def can_receive_settlements(self):
        return self._get_property("canReceiveSettlements")

    @property
    def organization(self):
        """Retrieve organization for an onboarding."""
        from .organization import Organization

        url = self._get_link("organization")

        if url:
            resp = self.client.organizations.perform_api_call(self.client.organizations.REST_READ, url)
            return Organization(resp)

    def is_needs_data(self):
        return self.status == self.STATUS_NEEDS_DATA

    def is_in_review(self):
        return self.status == self.STATUS_IN_REVIEW

    def is_completed(self):
        return self.status == self.STATUS_COMPLETED



from ..objects.list import List
from ..objects.method import Method
from .base import Base


class Methods(Base):
    def get_resource_object(self, result):
        return Method(result)

    def all(self, **params):
        """List all mollie payment methods, including methods that aren't activated in your profile."""
        path = "methods/all"
        result = self.perform_api_call(self.REST_LIST, path, params=params)
        return List(result, self.get_resource_object({}).__class__, self.client)



from ..error import IdentifierError
from ..objects.subscription import Subscription
from .base import Base


class CustomerSubscriptions(Base):
    RESOURCE_ID_PREFIX = "sub_"
    customer_id = None

    def get_resource_object(self, result):
        return Subscription(result, self.client)

    def get(self, subscription_id, **params):
        if not subscription_id or not subscription_id.startswith(self.RESOURCE_ID_PREFIX):
            raise IdentifierError(
                f"Invalid subscription ID: '{subscription_id}'. A subscription ID "
                f"should start with '{self.RESOURCE_ID_PREFIX}'."
            )
        return super().get(subscription_id, **params)

    def delete(self, subscription_id, data=None):
        """Cancel subscription and return the subscription object.

        Deleting a subscription causes the subscription status to changed to 'canceled'.
        The updated subscription object is returned.
        """
        if not subscription_id or not subscription_id.startswith(self.RESOURCE_ID_PREFIX):
            raise IdentifierError(
                f"Invalid subscription ID: '{subscription_id}'. A subscription ID "
                f"should start with '{self.RESOURCE_ID_PREFIX}'."
            )
        result = super().delete(subscription_id, data)
        return self.get_resource_object(result)

    def get_resource_name(self):
        return f"customers/{self.customer_id}/subscriptions"

    def with_parent_id(self, customer_id):
        self.customer_id = customer_id
        return self

    def on(self, customer):
        return self.with_parent_id(customer.id)



from .chargebacks import Chargebacks


class ProfileChargebacks(Chargebacks):
    profile_id = None

    def get_resource_name(self):
        return f"chargebacks?profileId={self.profile_id}"

    def with_parent_id(self, profile_id):
        self.profile_id = profile_id
        return self

    def on(self, profile):
        return self.with_parent_id(profile.id)



from ..error import RequestError
from .methods import Methods


class ProfileMethods(Methods):
    RESOURCE_REQUIRED_METHODS = ["giftcard", "voucher"]

    profile_id = None
    method_id = None
    resource_id = None

    def get_resource_name(self):
        path = f"profiles/{self.profile_id}/methods/{self.method_id}"
        if self.method_id in self.RESOURCE_REQUIRED_METHODS:
            path += f"/issuers/{self.resource_id}"
        return path

    def list(self, **params):
        return self.client.methods.list(profileId=self.profile_id)

    def delete(self, resource_id=None, *args, **kwargs):
        if self.method_id in self.RESOURCE_REQUIRED_METHODS and resource_id is None:
            raise RequestError(f"resource_id is required when disabling a {self.method_id}")
        self.resource_id = resource_id
        path = self.get_resource_name()
        return self.perform_api_call(self.REST_DELETE, path, None)

    def create(self, resource_id=None, data=None, **params):
        if self.method_id in self.RESOURCE_REQUIRED_METHODS and resource_id is None:
            raise RequestError(f"resource_id is required when enabling a {self.method_id}")
        self.resource_id = resource_id
        path = self.get_resource_name()
        result = self.perform_api_call(self.REST_CREATE, path, data=data, params=params)
        return self.get_resource_object(result)

    def with_parent_id(self, profile_id, method=None):
        self.method_id = method
        self.profile_id = profile_id
        return self

    def on(self, profile, method=None):
        return self.with_parent_id(profile.id, method)



from ..error import IdentifierError
from ..objects.payment import Payment
from .base import Base


class Payments(Base):
    RESOURCE_ID_PREFIX = "tr_"

    def get_resource_object(self, result):
        return Payment(result, self.client)

    def get(self, payment_id, **params):
        if not payment_id or not payment_id.startswith(self.RESOURCE_ID_PREFIX):
            raise IdentifierError(
                f"Invalid payment ID: '{payment_id}'. A payment ID should start with '{self.RESOURCE_ID_PREFIX}'."
            )
        return super().get(payment_id, **params)

    def delete(self, payment_id, data=None):
        """Cancel payment and return the payment object.

        Deleting a payment causes the payment status to change to canceled.
        The updated payment object is returned.
        """
        if not payment_id or not payment_id.startswith(self.RESOURCE_ID_PREFIX):
            raise IdentifierError(
                f"Invalid payment ID: '{payment_id}'. A payment ID should start with '{self.RESOURCE_ID_PREFIX}'."
            )
        result = super().delete(payment_id, data)
        return self.get_resource_object(result)



from ..error import IdentifierError
from ..objects.organization import Organization
from .base import Base


class Organizations(Base):
    RESOURCE_ID_PREFIX = "org_"

    def get_resource_object(self, result):
        return Organization(result, self.client)

    def get(self, organization_id, **params):
        if not organization_id or (
            not organization_id.startswith(self.RESOURCE_ID_PREFIX) and not organization_id == "me"
        ):
            raise IdentifierError(
                f"Invalid organization ID: '{organization_id}'. A organization ID should start "
                f"with '{self.RESOURCE_ID_PREFIX}' or it should be 'me'."
            )
        return super().get(organization_id, **params)



from .base import Base


class Customer(Base):
    @classmethod
    def get_resource_class(cls, client):
        from ..resources.customers import Customers

        return Customers(client)

    @property
    def id(self):
        return self._get_property("id")

    @property
    def name(self):
        return self._get_property("name")

    @property
    def email(self):
        return self._get_property("email")

    @property
    def locale(self):
        return self._get_property("locale")

    @property
    def metadata(self):
        return self._get_property("metadata")

    @property
    def mode(self):
        return self._get_property("mode")

    @property
    def resource(self):
        return self._get_property("resource")

    @property
    def created_at(self):
        return self._get_property("createdAt")

    @property
    def subscriptions(self):
        """Return the subscription list for the customer."""
        return self.client.customer_subscriptions.on(self).list()

    @property
    def mandates(self):
        """Return the mandate list for the customer."""
        return self.client.customer_mandates.on(self).list()

    @property
    def payments(self):
        """Return the payment list for the customer."""
        return self.client.customer_payments.on(self).list()
